# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Collect widget plugins together in a `Collection`.
"""

import inspect
import os
from os import path
from constants import TYPE_ERROR
from freezable import Freezable, freeze
from base import Widget, Page
from styles import Styles
import errors
from cStringIO import StringIO
import util


# FIXME: This function has no unit test
def find_modules_in_dir(src_dir):
    """
    Iterate through module names found in ``src_dir``.
    """
    if not (path.abspath(src_dir) == src_dir and path.isdir(src_dir)):
        return
    suffix = '.py'
    for name in sorted(os.listdir(src_dir)):
        if not name.endswith(suffix):
            continue
        pyfile = path.join(src_dir, name)
        if not path.isfile(pyfile):
            continue
        module = name[:-len(suffix)]
        if module == '__init__':
            continue
        yield (module, pyfile)


# FIXME: This function has no unit test
def iter_submodules(package, subpackage):
    """
    Import and iterate through all modules in ``package.subpackage``.
    """
    PACKAGE = __import__('%s.%s' % (package, subpackage))
    SUBPACKAGE = getattr(PACKAGE, subpackage)
    src_dir = path.dirname(path.abspath(SUBPACKAGE.__file__))
    for (module, pyfile) in find_modules_in_dir(src_dir):
        fullname = '%s.%s' % (SUBPACKAGE.__name__, module)
        MODULE = __import__(fullname)
        yield (module, getattr(SUBPACKAGE, module))


class Collection(Freezable):
    def __init__(self, name=None):
        self.__name = name
        self.__plugins = dict()

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.__name)

    def register(self, plugin, override=False):
        if not inspect.isclass(plugin):
            raise TypeError(
                'plugin must be a class; got %r' % (plugin,)
            )
        if not issubclass(plugin, Widget):
            raise ValueError(
                'plugin must be a Widget subclass; got %r' % plugin
            )
        name = plugin.__name__
        if name in self.__plugins:
            if not override:
                raise errors.OverrideError(
                    new=plugin, orig=self.__plugins[name]
                )
        else:
            if hasattr(self, name):
                raise errors.AttributeConflict(
                    plugin=plugin,
                    registrar=self.__class__.__name__,
                    attribute=name,
                )
            if override:
                raise errors.MissingOverrideError(name=name, plugin=plugin)
        self.__plugins[name] = plugin
        setattr(self, name, plugin)

    def register_builtins(self, override=False):
        for plugin in builtins():
            self.register(plugin, override=override)

    def __len__(self):
        """
        Return the number of available plugins.
        """
        return len(self.__plugins)

    def __iter__(self):
        """
        Iterate through the names of the available plugins.
        """
        for name in sorted(self.__plugins):
            yield name

    def __call__(self):
        for name in sorted(self.__plugins):
            yield self.__plugins[name]

    def __contains__(self, name):
        """
        Return ``True`` if a plugin named ``name`` has been registered.
        """
        return name in self.__plugins

    def __getitem__(self, name):
        """
        Return the plugin named ``name``.

        If no plugin named ``name`` has been registered, ``KeyError`` is raised.
        """
        return self.__plugins[name]

    def __todict__(self):
        """
        Return a copy of the private dict mapping plugin name to plugin.
        """
        return dict(self.__plugins)


builtins = Collection(name='builtins')


# FIXME: This function has no unit test
def load_builtins(package):
    """
    Load all builtin plugins.
    """
    for (name, module) in iter_submodules(package, 'plugins'):
        pass
    freeze(builtins)
