# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Form and form element plugins.
"""

from wehjit import base, builtins
from wehjit.state import *


class Form(base.Form):
    js_class = 'Form'

    javascript = """
    Wehjit.bases.Form = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            this.focused = null;
            $each(this.el.elements, function(field) {
                field.connect('focus', this);
            }, this);
            var parent = this.get_parent();
            if (parent && parent.klass == 'Dialog') {
                parent.addEvent('run', this.on_run.bind(this));
                this.parent = parent;
            }
            this.formdata = null;
        },

        on_focus: function(field, event) {
            this.focused = field;
        },

        on_run: function(dialog, params) {
            console.assert(dialog == this.parent);
            this.refocus();
        },

        refocus: function() {
            console.log('refocus', this.id, this.focused);
            if (this.focused) {
                this.focused.focus();
                return true;
            }
            if (this.el.elements.length > 0) {
                this.el.elements[0].focus();
                return true;
            }
            return false;
        },

        get_data: function(raw) {
            console.log('Form.get_data');
            var rawdata = this.el.get_data();
            if (raw) {
                return rawdata;
            }
            var data = {};

            if (this.formdata == null) {
                $each(rawdata, function(value, key) {
                    if (value !== '') {
                        data[key] = value;
                    }
                });
            }
            else {
                $each(rawdata, function(value, key) {
                    var old = this.formdata[key];
                    if (old  == undefined && value === '') {
                        return;
                    }
                    if (old != value) {
                        console.log('changed: %s = %s', key, value);
                        data[key] = value;
                    }
                }, this);
            }

            return data;
        },

        set_data: function(data) {
            console.log('Form.set_data', data);
            this.focused = null;
            if ($type(data) == 'object') {
                this.formdata = data;
            }
            else {
                this.formdata = null;
            }
            this.el.set_data(data);
        },

        reset: function() {
            this.formdata = null;
            this.focused = null;
            this.el.reset();
        },

    });
    """

builtins.register(Form)


class Fieldset(base.Container):
    """
    A fieldset to visually and semantically group form elements.
    """

    label = Dynamic('label')
    _label = 'Fieldset Label'

    xml = """
    <fieldset
        xmlns:py="http://genshi.edgewall.org/"
        class="${klass}"
        id="${id}"

    >
    <legend py:content="label" />
    <input py:for="child in children" py:replace="child.generate()" />
    </fieldset>
    """

    style_vars = (
        ('colorborder_fieldset', '#966'),
        ('colorbg_legend', '#966'),
        ('color_legend', '#fff'),
    )

    style = (
        ('', (
            ('border', '2px solid %(colorborder_fieldset)s'),
            ('-moz-border-radius', '0.75em'),
            ('margin', '2em 0'),
            ('width', '%(width_content)s'),
        )),
        ('> legend', (
            ('font-weight', 'bold'),
            ('letter-spacing', '0.15em'),
            ('word-spacing', '0.25em'),
            ('border', 'inherit'),
            ('background-color', '%(colorbg_legend)s'),
            ('color', '%(color_legend)s'),
            ('padding', '0 0.5em'),
        )),
    )

builtins.register(Fieldset)


class Button(base.Field):

    type = StaticProp(name='type', default='button')

    xml = """
    <button
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        name="${name}"
        class="${css_classes}"
        type="${type}"
        py:content="label"
    />
    """

builtins.register(Button)


class Submit(Button):
    _type = 'submit'
    _label = 'Submit'

builtins.register(Submit)


class Reset(Button):
    _type = 'reset'
    _label = 'Reset'

builtins.register(Reset)


class Text(base.Field):
    """
    A single-line text field.
    """

    size = Dynamic('size', default=30)
    maxlength = Dynamic('maxlength')

    type = StaticProp(name='type', default='text')

    xml = """
    <label
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    ${label}
    <input
            name="${name}"
            id="${id}.__field__"
            type="${type}"
            size="${size}"
            maxlength="${maxlength}"
        />
    </label>
    """

    style = (
        ('input', (
            ('padding', '%(padding_input)s'),
            ('border', 'none'),
        )),
    )

builtins.register(Text)


class Password(Text):
    """
    A password field.
    """

    _type = 'password'

builtins.register(Password)


class Checkbox(base.Field):
    """
    A checkbox.
    """

    xml = """
    <label
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes} associate"
    >
        <input name="${name}" type="checkbox" />
        ${label}
    </label>
    """

builtins.register(Checkbox)


class CheckboxRow(Checkbox):
    """
    A checkbox to be placed in a `FieldTable`.
    """

    xml = """
    <tr
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    <td class="fieldlabel">${label}</td>
    <td>
    <label class="associate">
    <input name="${name}" type="checkbox" />
    ${label}
    </label>
    </td>
    </tr>
    """

builtins.register(CheckboxRow)


class Select(base.Field):
    """
    A single-select dropdown.
    """

    options = Dynamic('options', default=(('1', 'Yes'), ('0', 'No')))

    xml = """
    <label
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
        ${label}
        <select name="${name}" id="${id}.__field__">
            <option
                py:for="(optvalue,optlabel) in options"
                py:content="optlabel"
                value="${optvalue}"
            />
        </select>
    </label>
    """

builtins.register(Select)


class RadioField(base.Field):
    options = Dynamic('options', default=(('1', 'Yes'), ('0', 'No')))

    xml = """
    <fieldset
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
    >
        <legend py:content="label" />
        <label class="associate" py:for="(optvalue,optlabel) in options">
            <input type="radio" name="${name}" value="${optvalue}" />
            ${optlabel}
        </label>
    </fieldset>
    """

    style = (
        ('', (
            ('display', 'inline'),
        )),
    )

builtins.register(RadioField)
