# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
State management.
"""

import threading
import inspect
from freezable import Freezable, freeze
from constants import TYPE_ERROR, CALLABLE_ERROR, SET_ERROR, OVERRIDE_ERROR

__all__ = ['Static', 'SetOnce', 'Dynamic', 'StaticProp', 'DynamicProp']


def nice_repr(kind, value):
    if kind is callable and callable(value):
        return value.__name__
    return repr(value)


class Descriptor(object):
    """
    Base class for `StateBase` and `PropBase`.
    """

    params = (
        ('name', str, None),
        ('default', None, None),
        ('doc', str, None),
        ('json', bool, False),
    )
    args = ('name',)

    def __init__(self, *args, **kw):
        self.__repr = None
        self.__kw = kw
        if len(args) > len(self.args):
            raise TypeError(
                '%s.__init__() takes at most %d args, got %d' % (
                    self.__class__.__name__, len(self.args), len(args)
                )
            )
        for (i, arg) in enumerate(args):
            kw[self.args[i]] = arg
        for (key, kind, default) in self.params:
            value = kw.get(key, default)
            assert not hasattr(self, key), '%s.%s' % (self.__class__.__name__, key)
            setattr(self, key, value)
            if kind is None:
                continue
            if value is None and default is None and key != 'name':
                continue
            if kind is callable:
                if not callable(value):
                    nice = '.'.join([self.__class__.__name__, key])
                    raise TypeError(
                        CALLABLE_ERROR % (nice, type(value), value)
                    )
            elif not isinstance(value, kind):
                nice = '.'.join([self.__class__.__name__, key])
                raise TypeError(
                    TYPE_ERROR % (nice, kind, type(value), value)
                )
        self.pname = '_' + self.name

    def __repr__(self):
        if self.__repr is None:
            self.__repr = '%s(%s)' % (
                self.__class__.__name__, ', '.join(self._repr_iter())
            )
        return self.__repr

    def _repr_iter(self):
        d = dict(
            (key, (kind, default)) for (key, kind, default) in self.params
        )
        for key in self.args:
            (kind, default) = d[key]
            value = self.__kw[key]
            yield nice_repr(kind, value)
        for key in sorted(self.__kw):
            if key in self.args:
                continue
            (kind, default) = d[key]
            value = self.__kw[key]
            if value is not default:
                yield '%s=%s' % (key, nice_repr(kind, value))






class StateBase(Descriptor):
    """
    Base class for `Static` and `Dynamic` descriptor classes.
    """

    params = Descriptor.params + (
        ('kind', type, None),
        ('validator', callable, None),
    )

    def __set__(self, obj, value):
        if not (self.kind is None or isinstance(value, self.kind)):
            nice = '.'.join([obj.__class__.__name__, self.name])
            raise TypeError(
                TYPE_ERROR % (nice, self.kind, type(value), value)
            )
        if self.validator is not None:
            value = self.validator(obj, value)
        self._set(obj, value)


class Static(StateBase):
    """
    Descriptor for attributes that are read-only after the object is frozen.

    Attributes using the `Static` descriptor can be set via a kwarg passed to
    `Statefull.__init__()` or set directly via the attribute.

    For example:

    >>> class Foo(Statefull):
    ...     foo = Static('foo')
    ...
    >>> foo = Foo(foo='Hello world!')
    >>> foo.foo
    'Hello world!'
    >>> foo.foo = 'How are you?'
    >>> foo.foo
    'How are you?'

    You can also supply a default value when you create your `Static`
    descriptor.  For example

    >>> class Bar(Statefull):
    ...     bar = Static('bar', default='Raise it')
    ...
    >>> bar = Bar()
    >>> bar.bar
    'Raise it'
    >>> bar.bar = 'bar none'
    >>> bar.bar
    'bar none'

    But a kwarg will still override the default value:

    >>> bar2 = Bar(bar='the entry')
    >>> bar2.bar
    'the entry'

    A subclass can override the descriptor default without creating a new
    descriptor by setting the ``'_' + name`` class attribute.  For example:

    >>> class Baz(Bar):
    ...     _bar = 'looks more like baz to me'
    ...
    >>> baz = Baz()
    >>> baz.bar
    'looks more like baz to me'
    >>> baz.bar = 'no, looks like baz.bar'
    >>> baz.bar
    'no, looks like baz.bar'

    If you try to set a `Static` value after the `Statefull` instance is frozen,
    an ``AttributeError`` will be raised.  For example:

    >>> freeze(baz)  # doctest:+ELLIPSIS
    <wehjit.state.Baz object at 0x...>
    >>> baz.bar = 'wont work'
    Traceback (most recent call last):
      ...
    AttributeError: frozen: cannot set Baz.bar to 'wont work'
    """

    def __get__(self, obj, type_=None):
        if obj is None:
            return self
        return getattr(obj, self.pname, self.default)

    def _set(self, obj, value):
        if obj.__frozen__:
            raise AttributeError(
                SET_ERROR % (obj.__class__.__name__, self.name, value)
            )
        setattr(obj, self.pname, value)


class SetOnce(Static):
    def _set(self, obj, value):
        if obj.__frozen__:
            raise AttributeError(
                SET_ERROR % (obj.__class__.__name__, self.name, value)
            )
        orig = getattr(obj, self.pname, None)
        if orig is not None:
            raise AttributeError (
                OVERRIDE_ERROR % (obj.__class__.__name__, self.name, orig, value)
            )
        setattr(obj, self.pname, value)


class Dynamic(StateBase):
    """
    Descriptor for attributes that can have a per-request threadlocal value.

    `Dynamic` attributes are similar to `Static` attributes except that the
    value can be set even after the `Statefull` instance is frozen.  For
    example:

    >>> class Foo(Statefull):
    ...     bar = Dynamic('bar')
    ...
    >>> foo = Foo(bar='Hello world!')
    >>> freeze(foo)  # doctest:+ELLIPSIS
    <wehjit.state.Foo object at 0x...>
    >>> foo.bar = 'How are you?'
    >>> foo.bar
    'How are you?'

    When setting a `Dynamic` attribute after the `Statefull` instance is frozen,
    the value is stored in the threadlocal ``__request__`` dictionary:

    >>> foo.__request__
    {'bar': 'How are you?'}

    In practice it's a bad idea to manipulate ``__request__`` directly, but we
    will to demonstrate how things work.  If we changed the ``'bar'`` key in
    ``__request__``, it will change the value returned when we evaluate
    ``foo.bar``:

    >>> foo.__request__['bar'] = 'try this for a change'
    >>> foo.bar
    'try this for a change'

    Calling `Statefull.clear()` will remove all values from ``__request__``:

    >>> foo.clear()
    >>> foo.__request__
    {}

    When ``'bar'`` isn't in ``__request__``, evaluating ``foo.bar`` will
    retrieve whatever the value was when ``foo`` was frozen:

    >>> foo.bar
    'Hello world!'

    See `Static` more details about the behavior of `Dynamic` attributes
    *before* the `Statefull` instance is frozen.
    """
    def __get__(self, obj, type_=None):
        if obj is None:
            return self
        if obj.__frozen__:
            return obj.__request__.get(
                self.name,
                getattr(obj, self.pname, self.default)
            )
        return getattr(obj, self.pname, self.default)

    def __set__(self, obj, value):
        if self.validator is not None:
            value = self.validator(obj, value)
        if obj.__frozen__:
            obj.__request__[self.name] = value
        else:
            setattr(obj, self.pname, value)


class PropBase(Descriptor):
    """
    Base class for `StaticProp` and `DynamicProp` descriptor classes.
    """

    params = Descriptor.params + (
        ('get', callable, None),
    )
    args = ('get', 'name')

    def __init__(self, get=None, name=None, **kw):
        if get is None and name is None:
            raise TypeError(
                '%s: must provide at least `get` or `name`' % (
                    self.__class__.__name__
                )
            )
        if not (get is None or kw.get('default') is None):
            raise TypeError(
                '%s: cannot provide both get=%r and default=%r' % (
                    self.__class__.__name__, get, kw['default']
                )
            )
        if callable(get):
            if name is None:
                name = get.__name__
            if kw.get('doc') is None:
                kw['doc'] = inspect.getdoc(get)
        super(PropBase, self).__init__(get, name, **kw)

    def __get__(self, obj, type_=None):
        if obj is None:
            return self
        if self.get is None:
            return getattr(obj, self.pname, self.default)
        return self.get(obj)

    def __set__(self, obj, value):
        raise AttributeError('read-only %s: cannot set %s.%s to %r' % (
                self.__class__.__name__, obj.__class__.__name__,
                self.name, value
            )
        )

    def __delete__(self, obj):
        raise AttributeError('read-only %s: cannot del %s.%s' % (
                self.__class__.__name__, obj.__class__.__name__, self.name
            )
        )


# FIXME: in case a StaticProp is expensive to compute but needed each request
# via the attribute (instead of __state__), Statefull should store the computed
# value on a special attribute like '_frozen_' + name, and StaticProp.__get__()
# should retrieve it there when obj is frozen.
class StaticProp(PropBase):
    pass


class DynamicProp(PropBase):
    pass


descriptors = (Static, Dynamic, StaticProp, DynamicProp)


class MetaClass(type):
    """
    A metaclass used to scan a new class for descriptors.

    See `Statefull`.
    """

    def __new__(meta, name, bases, dict):
        cls = type.__new__(meta, name, bases, dict)
        pairs = tuple(
            (desc, list()) for desc in descriptors
        )
        json = []
        for name in dir(cls):
            if name.startswith('_'):
                continue
            attr = getattr(cls, name)
            if not isinstance(attr, Descriptor):
                continue
            for (desc, accum) in pairs:
                if isinstance(attr, desc):
                    assert attr.name == name, '%s.%s: %r' % (
                        cls.__name__, name, attr
                    )
                    accum.append(name)
                    if attr.json:
                        json.append(attr)
                    break
        for (desc, accum) in pairs:
            setattr(cls, desc.__name__, tuple(sorted(accum)))
        cls.KwargKeys = frozenset(cls.Static + cls.Dynamic)
        cls.RequestKeys = frozenset(cls.Dynamic)
        cls.StateKeys = tuple(
            sorted(cls.Static + cls.Dynamic + cls.StaticProp)
        )
        cls.JSONKeys = tuple(sorted(
            filter(
                lambda k: getattr(cls, k).json,
                cls.StateKeys + cls.DynamicProp
            )
        ))
        if callable(getattr(cls, '_compile', None)):
            cls._compile()
        return cls


class Statefull(Freezable):
    __metaclass__ = MetaClass

    def __init__(self, **kw):
        self.__state__ = None
        self.__dynamic__ = None
        self.__threadlocal = threading.local()
        self.__request__ = self.__threadlocal.__dict__
        if not self.KwargKeys.issuperset(kw):
            extra = frozenset(kw) - self.KwargKeys
            raise TypeError(
                '%s.__init__() takes no such kwargs: %s' % (
                    self.klass,
                    ', '.join(repr(k) for k in sorted(extra))
                )
            )
        for (key, value) in kw.iteritems():
            setattr(self, key, value)

    def __setattr__(self, name, value):
        """
        If unfrozen, set attribute named ``name`` to ``value``.

        If this instance is frozen, an AttributeError will be raised.

        :param name: Name of attribute to set.
        :param value: Value to assign to attribute.
        """
        if self.__frozen__ and name not in self.RequestKeys:
            raise AttributeError(
                SET_ERROR % (self.__class__.__name__, name, value)
            )
        return object.__setattr__(self, name, value)

    def __freeze__(self):
        self.__state__ = tuple(self._state_iter())
        super(Statefull, self).__freeze__()

    def _state_iter(self):
        for key in self.StateKeys:
            yield (key, getattr(self, key))

    def state(self):
        """
        Return the current state, including any thread-local request state.
        """
        if self.__state__ is None:
            state = dict(self._state_iter())
        else:
            state = dict(self.__state__)
        if self.__frozen__ and self.__request__:
            state.update(self.__request__)
        for key in self.DynamicProp:
            state[key] = getattr(self, key)
        return state

    def request(self):
        """
        Return the `Dynamic` state, possibly overriden by thread-local values.
        """
        if self.__dynamic__ is None:
            request = dict(self._dynamic_iter())
        else:
            request = dict(self.__dynamic__)
        if self.__request__:
            request.update(self.__request__)
        return request

    def _dynamic_iter(self):
        for key in self.Dynamic:
            yield (key, getattr(self, key))

    def update_request(self, request):
        """
        Merge reqest into the possibly already existing request state.
        """
        if not self.__frozen__:
            return
        if isinstance(request, dict):
            request = request.iteritems()
        for (key, value) in request:
            if key in self.RequestKeys:
                self.__request__[key] = value

    def clear(self):
        """
        Clear the request state.
        """
        self.__request__.clear()
